import typer

from anaconda_cloud_auth.client import BaseClient
from anaconda_cloud_auth.console import console

app = typer.Typer(add_completion=False, help="Anaconda.cloud auth commands")


@app.command(name="info")
def auth_info() -> None:
    """Display information about the currently signed-in user"""
    client = BaseClient()
    response = client.get("/api/account")
    console.print("Your Anaconda Cloud info:")
    console.print(response.json())



class AuthenticationError(Exception):
    pass


class InvalidTokenError(AuthenticationError):
    pass


class TokenNotFoundError(Exception):
    pass


class LoginRequiredError(Exception):
    pass


class TokenExpiredError(Exception):
    pass



from typing import Any
from typing import Optional
from typing import Union
from urllib.parse import urljoin

import requests
from requests import PreparedRequest
from requests import Response
from requests.auth import AuthBase

from anaconda_cloud_auth import __version__ as version
from anaconda_cloud_auth.config import APIConfig
from anaconda_cloud_auth.config import AuthConfig
from anaconda_cloud_auth.exceptions import LoginRequiredError
from anaconda_cloud_auth.exceptions import TokenNotFoundError
from anaconda_cloud_auth.token import TokenInfo


class BearerAuth(AuthBase):
    def __init__(
        self, domain: Optional[str] = None, api_key: Optional[str] = None
    ) -> None:
        self.api_key = api_key
        if domain is None:
            domain = AuthConfig().domain

        self._token_info = TokenInfo(domain=domain)

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        if not self.api_key:
            try:
                r.headers[
                    "Authorization"
                ] = f"Bearer {self._token_info.get_access_token()}"
            except TokenNotFoundError:
                pass
        else:
            r.headers["Authorization"] = f"Bearer {self.api_key}"
        return r


class BaseClient(requests.Session):
    _user_agent: str = f"anaconda-cloud-auth/{version}"

    def __init__(
        self,
        domain: Optional[str] = None,
        api_key: Optional[str] = None,
        user_agent: Optional[str] = None,
    ):
        super().__init__()

        kwargs = {}
        if domain is not None:
            kwargs["domain"] = domain
        if api_key is not None:
            kwargs["key"] = api_key

        self.config = APIConfig(**kwargs)
        self._base_url = f"https://{self.config.domain}"
        self.headers["User-Agent"] = user_agent or self._user_agent
        self.auth = BearerAuth(api_key=self.config.key)

    def request(
        self,
        method: Union[str, bytes],
        url: Union[str, bytes],
        *args: Any,
        **kwargs: Any,
    ) -> Response:
        joined_url = urljoin(self._base_url, str(url))
        response = super().request(method, joined_url, *args, **kwargs)
        if response.status_code == 401 or response.status_code == 403:
            if response.request.headers.get("Authorization") is None:
                raise LoginRequiredError(
                    f"{response.reason}: You must login before using this API endpoint using\n"
                    f"  anaconda login"
                )
        return response


def client_factory(user_agent: Optional[str]) -> BaseClient:
    return BaseClient(user_agent=user_agent)



import logging
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler
from http.server import HTTPServer
from socket import socket
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from urllib.parse import parse_qs
from urllib.parse import urlparse

from pydantic.main import BaseModel

from anaconda_cloud_auth.exceptions import AuthenticationError

logger = logging.getLogger(__name__)


class Result(BaseModel):
    """This class is needed to capture the auth code redirect data"""

    auth_code: str = ""
    state: str = ""
    scopes: List[str] = []


TRequest = Union[socket, Tuple[bytes, socket]]


class AuthCodeRedirectServer(HTTPServer):
    """A simple http server to handle the incoming auth code redirect from Ory"""

    def __init__(self, oidc_path: str, server_address: Tuple[str, int]):
        super().__init__(server_address, AuthCodeRedirectRequestHandler)
        self.result: Union[Result, None] = None
        self.host_name = str(self.server_address[0])
        self.oidc_path = oidc_path

    def finish_request(self, request: TRequest, client_address: str) -> None:
        """Finish one request by instantiating RequestHandlerClass."""
        AuthCodeRedirectRequestHandler(
            self.oidc_path,
            self.host_name,
            request,
            client_address,
            server=self,
        )


class AuthCodeRedirectRequestHandler(BaseHTTPRequestHandler):
    """Request handler to get the auth code from the redirect from Ory"""

    server: AuthCodeRedirectServer

    def __init__(
        self,
        oidc_path: str,
        host_name: str,
        *args: Any,
        **kwargs: Any,
    ):
        # these are set before __init__ because __init__ calls the do_GET method
        self.oidc_path = oidc_path
        self.host_name = host_name

        super().__init__(*args, **kwargs)

    def log_message(self, format: str, *args: Any) -> None:
        """Override base method to suppress log message."""

    def _handle_auth(self, query_params: Dict[str, List[str]]) -> None:
        if "code" in query_params and "state" in query_params:
            location = "https://anaconda.cloud/local-login-success"
            self.server.result = Result(
                auth_code=query_params["code"][0],
                state=query_params["state"][0],
                scopes=query_params.get("scope", []),
            )
        else:
            location = "https://anaconda.cloud/local-login-error"

        self.send_response(HTTPStatus.TEMPORARY_REDIRECT)
        self.send_header("Location", location)
        self.end_headers()

    def _not_found(self) -> None:
        self.send_response(HTTPStatus.NOT_FOUND)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def do_GET(self) -> None:
        parsed_url = urlparse(f"http://{self.host_name}{self.path}")
        query_params = parse_qs(parsed_url.query)

        # Only accept requests to self.oidc_path
        if parsed_url.path != self.oidc_path:
            self._not_found()
        else:
            self._handle_auth(query_params)


def capture_auth_code(redirect_uri: str, state: str) -> str:
    parsed_url = urlparse(redirect_uri)

    host_name, port = parsed_url.netloc.split(":")
    server_port = int(port or "80")
    oidc_path = parsed_url.path

    logger.debug(f"Listening on: {redirect_uri}")

    with AuthCodeRedirectServer(oidc_path, (host_name, server_port)) as web_server:
        web_server.handle_request()

    result = web_server.result

    if result is None:
        raise AuthenticationError("Could not complete authentication")

    if result.auth_code is None:
        raise AuthenticationError("No authorization code")

    if result.state != state:
        raise AuthenticationError("State does not match")

    return result.auth_code



from getpass import getpass
from typing import Any
from typing import Protocol


class TConsole(Protocol):
    @staticmethod
    def print(*args: Any, **kwargs: Any) -> None:
        ...

    @staticmethod
    def input(*args: Any, **kwargs: Any) -> Any:
        ...


class SimpleConsole:
    """
    A very simple console class to mimic the necessary methods we use from rich,
    in case anaconda_cloud_cli is unavailable.
    """

    @staticmethod
    def print(*args: Any, **kwargs: Any) -> None:
        print(*args, **kwargs)

    @staticmethod
    def input(*args: Any, password: bool = False, **kwargs: Any) -> Any:
        if password:
            return getpass(args[0])
        else:
            return input(args[0])


console: TConsole


try:
    from anaconda_cloud_cli import console  # type: ignore

except ImportError:
    console = SimpleConsole()
